// Copyright (C) 2023 WYATT GROUP
// Please see the AUTHORS file for details.
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program. If not, see <https://www.gnu.org/licenses/>.

import 'dart:async';

import 'package:equatable/equatable.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:wyatt_architecture/wyatt_architecture.dart';
import 'package:wyatt_authentication_bloc/src/core/enums/authentication_status.dart';
import 'package:wyatt_authentication_bloc/src/core/utils/custom_routine.dart';
import 'package:wyatt_authentication_bloc/src/domain/entities/entities.dart';
import 'package:wyatt_authentication_bloc/src/domain/repositories/authentication_repository.dart';
import 'package:wyatt_type_utils/wyatt_type_utils.dart';

part 'authentication_state.dart';

/// Abstract authentication cubit class needs to be implemented in application.
///
/// This cubit is in charge of managing the global authentication state of
/// the application.
///
/// Its here you can override every callbacks and add your custom logic.
abstract class AuthenticationCubit<Data>
    extends Cubit<AuthenticationState<Data>> {
  AuthenticationCubit({
    required this.authenticationRepository,
  }) : super(const AuthenticationState.unknown()) {
    _listenForAuthenticationChanges();
  }
  final AuthenticationRepository<Data> authenticationRepository;

  SessionWrapper<Data>? _latestSession;

  void _listenForAuthenticationChanges() {
    authenticationRepository.sessionStream().asyncMap((wrapper) async {
      final event = wrapper.event;
      if (event is SignedInFromCacheEvent) {
        final customRoutineResult = await onSignInFromCache(wrapper);

        if (customRoutineResult.isOk) {
          final account = event.account;
          final sessionData = customRoutineResult.ok;

          final signedInSession = SessionWrapper(
            event: SignedInFromCacheEvent(account: account),
            session: Session<Data>(
              account: account,
              data: sessionData,
            ),
          );

          return signedInSession;
        }
      }
      return wrapper;
    }).listen((wrapper) async {
      _latestSession = wrapper;
      final session = wrapper.session;
      if (session != null) {
        emit(AuthenticationState<Data>.authenticated(wrapper));
        return;
      }
      emit(AuthenticationState<Data>.unauthenticated());
      return;
    });
  }

  /// {@macro refresh}
  FutureOr<void> refresh() async => CustomRoutine<Account, Data?>(
      routine: authenticationRepository.refresh,
      attachedLogic: onRefresh,
      onError: addError,
      onSuccess: (result, data) => authenticationRepository.addSession(
        SessionWrapper(
          event: RefreshedEvent(account: result),
          session: Session<Data>(
            account: result,
            data: data,
          ),
        ),
      ),
    ).call();

  /// {@macro reauthenticate}
  FutureOr<void> reauthenticate() async => CustomRoutine<Account, Data?>(
        routine: authenticationRepository.reauthenticate,
        attachedLogic: onReauthenticate,
        onError: addError,
        onSuccess: (result, data) => authenticationRepository.addSession(
          SessionWrapper(
            event: ReauthenticatedEvent(account: result),
            session: Session<Data>(
              account: result,
              data: data,
            ),
          ),
        ),
      ).call();

  /// {@macro signout}
  FutureOr<void> signOut() async => CustomRoutine<void, void>(
        routine: authenticationRepository.signOut,
        attachedLogic: (routineResult) => onSignOut(),
        onError: addError,
        onSuccess: (result, data) => authenticationRepository
            .addSession(SessionWrapper<Data>(event: const SignedOutEvent())),
      ).call();

  /// {@macro delete}
  FutureOr<void> delete() async => CustomRoutine<void, void>(
        routine: authenticationRepository.delete,
        attachedLogic: (routineResult) => onDelete(),
        onError: addError,
        onSuccess: (result, data) => authenticationRepository
            .addSession(SessionWrapper<Data>(event: const DeletedEvent())),
      ).call();

  /// Returns latest session wrapper.
  ///
  /// Contains latest event and latest session data (account + extra data)
  SessionWrapper<Data>? currentSession() => _latestSession;

  /// This callback is triggered when the user is automaticcaly logged in from
  /// the cache.
  ///
  /// For example: when the user is sign in from the Firebase cache.
  FutureOrResult<Data?> onSignInFromCache(SessionWrapper<Data> wrapper);

  /// This callback is triggered when the account is refreshed.
  ///
  /// For example: when the access token is refreshed.
  FutureOrResult<Data?> onRefresh(Result<Account, AppException> result);

  /// This callback is triggered when the account is re-authenticated
  ///
  /// For example: when the user is logged in and sign in
  /// from an another provider
  FutureOrResult<Data?> onReauthenticate(Result<Account, AppException> result);

  /// This callback is triggered when the user is logged out.
  ///
  /// For example: when the user clicks on the logout button.
  FutureOrResult<void> onSignOut();

  /// This callback is triggered when the current account is deleted from
  /// the remote.
  ///
  /// For example: when the user wants to delete his account from Firebase
  FutureOrResult<void> onDelete();
}
